/*
    yaa4xm - Yet Another Avs4X264Mod

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    The code is based on follow programs.
    on AVS-input from x264 (http://www.videolan.org/developers/x264.html).
    on writing process from avs2yuv_bm2 (http://komisar.gin.by/tools/avs2yuv/).
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <io.h>
#include <fcntl.h>
#include <ctype.h>

#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#undef __declspec
#define __declspec(i)
#undef EXTERN_C
#include "avisynth_c.h"

#define BUFSIZE 131072
#define AVS_INTERFACE_25 2
#define DEFAULT_BIN_NAME "x264.exe"

#define MY_VERSION "0.0.3"

/* maximum size of the sequence of filters to try on non script files */
#define AVS_MAX_SEQUENCE 5

#define LOAD_AVS_FUNC(name, continue_on_fail) \
{\
    h->func.name = (void*)GetProcAddress( h->library, #name );\
    if( !continue_on_fail && !h->func.name )\
        goto fail;\
}

#define FAIL_IF_ERROR(cond, name, destination, ...) \
    if (cond) {print_log(name, LOG_ERROR, __VA_ARGS__); goto destination;}

typedef struct {
    AVS_Clip *clip;
    AVS_ScriptEnvironment *env;
    HMODULE library;
    /* declare function pointers for the utilized functions to be loaded without __declspec,
       as the avisynth header does not compensate for this type of usage */
    struct {
        const char *(__stdcall *avs_clip_get_error)( AVS_Clip *clip );
        AVS_ScriptEnvironment *(__stdcall *avs_create_script_environment)( int version );
        void (__stdcall *avs_delete_script_environment)( AVS_ScriptEnvironment *env );
        AVS_VideoFrame *(__stdcall *avs_get_frame)( AVS_Clip *clip, int n );
        int (__stdcall *avs_get_version)( AVS_Clip *clip );
        const AVS_VideoInfo *(__stdcall *avs_get_video_info)( AVS_Clip *clip );
        int (__stdcall *avs_function_exists)( AVS_ScriptEnvironment *env, const char *name );
        AVS_Value (__stdcall *avs_invoke)( AVS_ScriptEnvironment *env, const char *name,
            AVS_Value args, const char **arg_names );
        void (__stdcall *avs_release_clip)( AVS_Clip *clip );
        void (__stdcall *avs_release_value)( AVS_Value value );
        void (__stdcall *avs_release_video_frame)( AVS_VideoFrame *frame );
        AVS_Clip *(__stdcall *avs_take_clip)( AVS_Value, AVS_ScriptEnvironment *env );
    } func;
} avs_hnd_t;

typedef enum {
    LOG_INFO,
    LOG_WARNING,
    LOG_ERROR
} log_t;

/* load the library and functions we require from it */
static int avs_load_library( avs_hnd_t *h )
{
    h->library = LoadLibrary("avisynth");
    if(!h->library)
        return -1;
    LOAD_AVS_FUNC(avs_clip_get_error, 0);
    LOAD_AVS_FUNC(avs_create_script_environment, 0);
    LOAD_AVS_FUNC(avs_delete_script_environment, 1);
    LOAD_AVS_FUNC(avs_get_frame, 0);
    LOAD_AVS_FUNC(avs_get_version, 0);
    LOAD_AVS_FUNC(avs_get_video_info, 0);
    LOAD_AVS_FUNC(avs_function_exists, 0);
    LOAD_AVS_FUNC(avs_invoke, 0);
    LOAD_AVS_FUNC(avs_release_clip, 0);
    LOAD_AVS_FUNC(avs_release_value, 0);
    LOAD_AVS_FUNC(avs_release_video_frame, 0);
    LOAD_AVS_FUNC(avs_take_clip, 0);
    return 0;
fail:
    FreeLibrary(h->library);
    return -1;
}

static void print_log(const char *name, log_t level, const char *message, ...)
{
    char *prefix;
    va_list args;
    va_start(args, message);
    switch (level) {
        case LOG_INFO:
            prefix = "info";
            break;
        case LOG_WARNING:
            prefix = "warning";
            break;
        case LOG_ERROR:
            prefix = "error";
            break;
        default:
            prefix = "unknown";
            break;
    }
    fprintf(stderr, "%s [%s]: ", name, prefix);
    vfprintf(stderr, message, args);
    va_end(args);
}

static void usage(void)
{
    fprintf(stderr,
            "YAA4XM - Yet Another Avs4X264Mod version %s\n"
            "built on %s %s\n"
            "\n"
            "a tool to encode videos using Avisynth with x264 under Windows.\n"
            "\n"
            "usage: yaa4xm input.avs [--x264 <bin>] <x264 options>\n"
            "  --x264 : path of x264.exe (default \"x264.exe\")\n"
            "\n"
            "example:\n"
            "  yaa4xm input.avs --x264 C:\\foo\\bar\\x264_x64_10bit.exe --preset slower\n"
            "  --tune film --level 4 --input-depth 16 -o output.mkv\n",
            MY_VERSION, __DATE__, __TIME__);
}

int main(int argc, char **argv)
{
    int exitcode = 1;

    if (argc < 4) {
        usage();
        goto ret;
    }

/* parse options*/
    char *infile = NULL;
    char *x264_bin = DEFAULT_BIN_NAME;
    int has_output = 0;
    int is_interlaced = 0;
    int is_tcfile_used = 0;
    int is_highbit_depth = 0;
    char avs_ext[5] = "";

    for (int i = 1; i < argc; i++) {
        FAIL_IF_ERROR(!strcmp(argv[i], "-h") || !strcmp(argv[i], "--help") ||
                      !strcmp(argv[i], "--longhelp") || !strcmp(argv[i], "--fullhelp"),
                      "yaa4xm", ret, "Invarid argument \"%s\".\n", argv[i]);
        FAIL_IF_ERROR(!strcmp(argv[i], "--frames") || !strcmp(argv[i], "--seek"),
                      "yaa4xm", ret, "This tool doesn't support --seek/--frames option.\n");
        if (!strcmp(argv[i], "-o") || !strcmp(argv[i], "--output")) {
            FAIL_IF_ERROR(i == argc - 1, "yaa4xm", ret, "No output target specified.\n");
            has_output = 1;
            i++;
            continue;
        }
        if (!strcmp(argv[i], "--tff") || !strcmp(argv[i], "--bff")) {
            is_interlaced = 1;
            continue;
        }
        if (!strcmp(argv[i], "--tcfile-in")) {
            is_tcfile_used = 1;
            i++;
            continue;
        }
        if (!strcmp(argv[i], "--input-depth")) {
            is_highbit_depth = atoi(argv[++i]) - 8;
            FAIL_IF_ERROR(is_highbit_depth < 0 || is_highbit_depth > 8, "yaa4xm",
                          ret, "Invarid bitdepth \"%s\" was specified", argv[i]);
            continue;
        }
        if (!strcmp(argv[i], "--x264")) {
            FAIL_IF_ERROR(i == argc - 1, "yaa4xm", ret, "--x264 needs an argument.\n");
            x264_bin = argv[i + 1];
            for (int j = i; j < argc - 2; j++)
                argv[j] = argv[j + 2];
            argc -= 2;
            i -= 1;
            continue;
        }
        int len = strlen(argv[i]);
        for (int j = len - 4; j < len; j++)
            avs_ext[j - (len - 4)] = tolower(argv[i][j]);
        if (!strcmp(avs_ext, ".avs")) {
            infile = argv[i];
            for (int k = i; k < argc - 1; k++)
                argv[k] = argv[k + 1];
            argc -= 1;
            i -= 1;
        }
    }

    FAIL_IF_ERROR(!infile, "yaa4xm", ret, "No avs script found.\n");
    FAIL_IF_ERROR(!has_output, "yaa4xm", ret, "No output target specified.\n");

/* avisynth processing*/
    avs_hnd_t avs_h = {0};
    FAIL_IF_ERROR(avs_load_library(&avs_h), "avs", ret, "failed to load avisynth\n");

    avs_h.env = avs_h.func.avs_create_script_environment(AVS_INTERFACE_25);
    FAIL_IF_ERROR(!avs_h.env, "avs", avs_close, "failed to initiate avisynth\n" );

    AVS_Value arg = avs_new_value_string(infile);
    AVS_Value res = avs_h.func.avs_invoke(avs_h.env, "Import", arg, NULL );
    FAIL_IF_ERROR(avs_is_error(res), "avs", avs_close, "%s\n", avs_as_string(res));

    AVS_Value mt_test = avs_h.func.avs_invoke( avs_h.env, "GetMTMode", avs_new_value_bool(0), NULL);
    int mt_mode = avs_is_int(mt_test) ? avs_as_int(mt_test) : 0;
    avs_h.func.avs_release_value(mt_test);
    if(mt_mode > 0 && mt_mode < 5) {
        AVS_Value temp = avs_h.func.avs_invoke(avs_h.env, "Distributor", res, NULL);
        avs_h.func.avs_release_value(res);
        res = temp;
    }

    FAIL_IF_ERROR(!avs_is_clip(res), "avs", avs_close,
                  "\"%s\" is not a clip\n", infile );

    avs_h.clip = avs_h.func.avs_take_clip(res, avs_h.env);

    const AVS_VideoInfo *vi = avs_h.func.avs_get_video_info( avs_h.clip );
    FAIL_IF_ERROR(!avs_has_video(vi), "avs", avs_close,
                  "\"%s\" has no video data\n", infile );

    FAIL_IF_ERROR(!avs_is_yv24(vi) && !avs_is_yv16(vi) && (vi->width & 1 || vi->height & 1),
                  "avs", avs_close, "input clip width or height not divisible by 2 (%dx%d)\n",
                  vi->width, vi->height );

    if (!avs_is_yv24(vi) && !avs_is_yv16(vi) && !avs_is_yv12(vi)) {
        avs_h.func.avs_release_clip(avs_h.clip);
        print_log("avs", LOG_WARNING, "Converting input clip to YV12\n");
        const char *arg_name[2] = {NULL, "interlaced"};
        AVS_Value arg_arr[2] = {res, avs_new_value_bool(is_interlaced)};
        AVS_Value res2 = avs_h.func.avs_invoke(avs_h.env, "ConvertToYV12",
                                               avs_new_value_array(arg_arr, 2),
                                               arg_name);
        FAIL_IF_ERROR(avs_is_error(res2), "avs", avs_close,
                      "Couldn't convert input clip to YV12\n");
        avs_h.clip = avs_h.func.avs_take_clip(res2, avs_h.env);
        avs_h.func.avs_release_value(res2);
        vi = avs_h.func.avs_get_video_info(avs_h.clip);
    }
    avs_h.func.avs_release_value(res);

/* generate commandline */
    char *raw_csp = avs_is_yv16(vi) ? "i422" :
                    avs_is_yv24(vi) ? "i444" :
                                      "i420";

    char fps_string[32] = "";
    if (!is_tcfile_used)
        sprintf(fps_string, "--fps %d/%d", vi->fps_numerator, vi->fps_denominator);

    char *cmd = NULL;
    FAIL_IF_ERROR(!(cmd = (char *)malloc(4096)), "yaa4xm", avs_close, "malloc failed\n")

    sprintf(cmd, "\"%s\" - --demuxer raw --input-csp %s --input-res %dx%d --frames %d %s",
            x264_bin, raw_csp, is_highbit_depth ? vi->width >> 1 : vi->width,
            vi->height, vi->num_frames, fps_string);
    for (int i = 1; i < argc; i++) {
        if (strchr(argv[i], ' '))
            sprintf(argv[i], "\"%s\"", argv[i]);
        sprintf(cmd, "%s %s", cmd, argv[i]);
    }

    print_log("yaa4xm", LOG_INFO, "%s\n", cmd);

/* create pipe */
    FILE *pipe_h = NULL;
    pipe_h = _popen(cmd, "wb");
    free(cmd);
    FAIL_IF_ERROR(!pipe_h, "yaa4xm", avs_close, "Failed to create pipe.\n");
    FAIL_IF_ERROR(setvbuf(pipe_h, NULL, _IOFBF, BUFSIZE), "yaa4xm", pipe_close,
                  "Failed to create buffer for pipe.\n");

/* writing */
    int chroma_h_shift = avs_is_yv24(vi) ? 0 : 1;
    int chroma_v_shift = avs_is_yv12(vi) ? 1 : 0;
    int write_target = (vi->width * vi->height * avs_bits_per_pixel(vi)) >> 3;

    for (int frame = 0; frame < vi->num_frames; frame++) {
        static const int planes[] = {AVS_PLANAR_Y, AVS_PLANAR_U, AVS_PLANAR_V};
        AVS_VideoFrame *frm = avs_h.func.avs_get_frame(avs_h.clip, frame);
        const char *err = avs_h.func.avs_clip_get_error(avs_h.clip);
        FAIL_IF_ERROR(err, "avs", pipe_close, "%s occurred while reading frame %d\n", err, frame );

        int wrote = 0;
        for(int p = 0; p < 3; p++) {
            int w = vi->width  >> (p ? chroma_h_shift : 0);
            int h = vi->height >> (p ? chroma_v_shift : 0);
            int pitch = avs_get_pitch_p(frm, planes[p]);
            const BYTE *data = avs_get_read_ptr_p(frm, planes[p]);
            for(int y = 0; y < h; y++) {
                wrote += fwrite(data, 1, w, pipe_h);
                data += pitch;
            }
        }

        avs_h.func.avs_release_video_frame(frm);
        FAIL_IF_ERROR(wrote != write_target, "yaa4xm", pipe_close,
                      "wrote only %d of %d bytes\n", wrote, write_target);
    }

    fflush(pipe_h);
    exitcode = 0;

pipe_close:
    _pclose(pipe_h);

avs_close:
    avs_h.func.avs_release_clip(avs_h.clip);
    if(avs_h.func.avs_delete_script_environment)
        avs_h.func.avs_delete_script_environment(avs_h.env);
    FreeLibrary(avs_h.library);

ret:
    return exitcode;
}
